CON
  _clkmode = XTAL1|PLL16X
  _xinfreq = 5_000_000
  
CON
  columns  = driver#res_x / 8
  rows     = driver#res_y / font#height
  bcnt     = columns * rows

  rows_raw = (driver#res_y + font#height - 1) / font#height
  bcnt_raw = columns * rows_raw

  dcolour  = driver#dcolour & $FFFF * $00010001
  
OBJ
  driver: "waitvid.50xH.driver.2048"
    font: "generic8x8-1font"
  
VAR
  long  link[driver#res_m]                              ' driver mailbox
  long  plte[bcnt_raw / 2]                              ' colour buffer (4n aligned)
  word  scrn[bcnt_raw / 2]                              ' screen buffer (2n aligned)

PUB selftest | x, y, c

  longfill(@plte{0}, dcolour, constant(bcnt_raw / 2))   ' default colour
  wordfill(@scrn{0},   $2020, constant(bcnt_raw / 2))   ' clear screen

  init                                                  ' start driver

  waitcnt(clkfreq*3 + cnt)
  redefine_char($20, @new_spc)                          ' new space character
  waitcnt(clkfreq*2 + cnt)
  redefine_char($20, @old_spc)                          ' reset

' setup plot functionality

  redefine_plot                                         ' 128..143

' draw some dots

  repeat
    repeat bcnt_raw * 5
      plot(||(?x//(columns*2)), ||(y?//(rows*2)), ?c)
    repeat bcnt_raw * 5
      plot(||(?x//(columns*2)), ||(y?//(rows*2)), -1)
    
DAT
new_spc byte    $FF, $81[6], $FF, $81[3], $FF, $81[3], $FF
old_spc byte    0[16]

PRI init

  link{0} := @scrn.byte[bcnt_raw - columns]             ' initial screen buffer and
  link[1] := font#height << 24 | font.addr              ' font definition
  link[2] := @plte{0}                                   ' colour buffer
  
  return driver.init(-1, @link{0})                      ' video driver and pixel generator

PRI redefine_char(c, buffer) | base

  base := font.addr + c.byte{0}                         ' start location
  repeat font#height
    byte[base] := byte[buffer++]
    base += 256
    
PRI redefine_plot | i, ptr

  ptr := font.addr 
  repeat i from 0 to constant(font#height/2 -1)
    bytemove(ptr + (i << 8) +                                  128,  @sg1, 16)
    bytemove(ptr + (i << 8) + constant(font#height / 2 * 256 + 128), @sg2, 16)
  
DAT
sg1     byte    $00, $0F, $F0, $FF, $00, $0F, $F0, $FF, $00, $0F, $F0, $FF, $00, $0F, $F0, $FF
sg2     byte    $00, $00, $00, $00, $0F, $0F, $0F, $0F, $F0, $F0, $F0, $F0, $FF, $FF, $FF, $FF

PRI plot(x, y, c) | pv, nv

  nv := |< (((y & 1) << 1) | (x & 1))   
  x >>= 1
  y >>= 1
  pv := c_get(x, y) & $0F
  if c == -1
    nv := pv & !nv
  else
    nv := pv |  nv
  c_set(x, y, nv|128)    

' finally set colour

  plte.byte[(y * columns + x) << 1 + 1] := c & driver#hv_mask | driver#hv_idle
  
PRI c_set(x, y, c)

  scrn.byte[bcnt_raw - (y * columns) - ++x] := c

PRI c_get(x, y) : c

  c := scrn.byte[bcnt_raw - (y * columns) - ++x]

DAT